#include <fcntl.h>
#include <stdio.h>
#include <stdlib.h>
#include <unistd.h>
#include <linux/i2c-dev.h>
#include <sys/ioctl.h>
#include <stdint.h>
#include "smbus.h"
#include <getopt.h>

#define WEATHER_I2C_BUS         "/dev/i2c-2"
#define WEATHER_SLAVE_ADDR      0x76
#define WEATHER_READ_ID		0xD0
#define	REG_DATA		0xF7
#define REG_CONTROL 		0xF4
#define REG_CONFIG		0xF5
#define REG_CONTROL_HUM		0xF2
#define REG_HUM_MSB		0xFD
#define REG_HUM_LSB		0xFE
#define OVERSAMPLE_TEMP		2
#define OVERSAMPLE_PRES		2
#define MODE			1
#define OVERSAMPLE_HUM		2

typedef struct calibration
{
	unsigned short	dig_T1;
	signed short	dig_T2;
	signed short	dig_T3;
	unsigned short 	dig_P1;
	signed short	dig_P2;	
	signed short	dig_P3;
	signed short	dig_P4;
	signed short	dig_P5;
	signed short	dig_P6;
	signed short	dig_P7;
	signed short	dig_P8;
	signed short	dig_P9;
	unsigned char 	dig_H1;
	signed short 	dig_H2;
	unsigned char 	dig_H3;
	signed short 	dig_H4;
	signed short 	dig_H5;
	signed char 	dig_H6;
} calibration;

char raw_data[8];
calibration cal;
int wait_time;
signed int t_fine;

int openDevice()
{
	int fd = 0;
	
	fd = open(WEATHER_I2C_BUS, O_RDWR);
	if (fd < 0) {
		perror(WEATHER_I2C_BUS);
		exit(EXIT_FAILURE);
	}
	
	if (ioctl(fd, I2C_SLAVE, WEATHER_SLAVE_ADDR) < 0) {
		perror("Slave unreachable");
		exit(EXIT_FAILURE);
	}

	return fd;
}

static void display_help(const char * name)
{
    	fprintf(stderr, "usage: %s options...\n", name);
	fprintf(stderr, " options:\n");
	fprintf(stderr, " -t --temperature\n");
	fprintf(stderr, " -h --humidity \n");
	fprintf(stderr, " -p --pressure\n");
	fprintf(stderr, " -h --help Prints this help\n\n");
	fprintf(stderr, "Example: %s --temperature --pressure\n\n", name);
}

unsigned int readRawTemperature()
{
	int fd = openDevice();
	sleep(wait_time/1000) ;
	
	// Read temperature/pressure/humidity
	i2c_smbus_read_i2c_block_data(fd, REG_DATA, 8, raw_data);
	unsigned int temp_raw = (raw_data[3] << 12) | (raw_data[4] << 4) | (raw_data[5] >> 4);
	close(fd);
	return temp_raw;
}

// Compensate Temperature : page 50
signed int compensateTemperature()
{
	unsigned temp_raw = readRawTemperature();
	signed int temperature, var1, var2;

	var1 = ((((temp_raw>>3)-(cal.dig_T1<<1)))*(cal.dig_T2)) >> 11;
	var2 = (((((temp_raw>>4) - (cal.dig_T1)) * ((temp_raw>>4) - (cal.dig_T1))) >> 12) * (cal.dig_T3)) >> 14;
	t_fine = var1 + var2;
	temperature = (t_fine * 5 + 128) >> 8;
}

double getTemperature()
{
	return (double)compensateTemperature() / 100;
}

unsigned int readRawPressure()
{
	int fd = openDevice();
	sleep(wait_time/1000);
	
	// Read temperature/pressure/humidity
	i2c_smbus_read_i2c_block_data(fd, REG_DATA, 8, raw_data);
	unsigned int pres_raw = (raw_data[0] << 12) | (raw_data[1] << 4) | (raw_data[2] >> 4);
	
	close(fd);
	return pres_raw;
}

// Compensate Pressure : page 49
unsigned int compensatePressure()
{
	unsigned int pres_raw = readRawPressure();
	double var1, var2, pressure;

	var1 = (double)t_fine / 2.0 - 64000.0;
	var2 = var1 * var1 * cal.dig_P6 / 32768.0;
	var2 = var2 + var1 * cal.dig_P5 * 2.0;
	var2 = var2 / 4.0 + cal.dig_P4 * 65536.0;
	var1 = (cal.dig_P3 * var1 * var1 / 524288.0 + cal.dig_P2 * var1) / 524288.0;
	var1 = (1.0 + var1 / 32768.0) * cal.dig_P1;
	  
	if (var1 == 0)
		pressure=0;
	else
	{
 	    pressure = 1048576.0 - pres_raw;
	    pressure = ((pressure - var2 / 4096.0) * 6250.0) / var1;
	    var1 = cal.dig_P9 * pressure * pressure / 2147483648.0;
	    var2 = pressure * cal.dig_P8 / 32768.0;
	    pressure = pressure + (var1 + var2 + cal.dig_P7) / 16.0;
 	}
 	
 	return pressure;
}

double getPressure()
{
	return ((double)compensatePressure()) / 100;
}

unsigned int readRawHumidity()
{
	int fd = openDevice();
	sleep(wait_time/1000);
	
	// Read temperature/pressure/humidity
	i2c_smbus_read_i2c_block_data(fd, REG_DATA, 8, raw_data);
	unsigned int hum_raw = (raw_data[6] << 8) | raw_data[7];	
	
	close(fd);
	return hum_raw;	
}

// Compensate Humidity : page 49
double compensateHumidity()
{
	signed int hum_raw = readRawHumidity();
	double humidity;
	
	humidity = (((double)t_fine) - 76800.0);	
	humidity = (hum_raw - (cal.dig_H4 * 64.0 + cal.dig_H5 / 16384.0 * humidity)) * (cal.dig_H2 / 65536.0 * (1.0 + cal.dig_H6 / 67108864.0 * humidity * (1.0 + cal.dig_H3 / 67108864.0 * humidity)));
	
	humidity = humidity * (1.0 - ((double)cal.dig_H1) * humidity / 524288.0);
	if (humidity > 100.0)
		humidity = 100.0;
	else if (humidity < 0.0)
		humidity = 0.0;
	return humidity;
}

double getHumidity()
{
	return (double)compensateHumidity();
}

void startCal()
{
	int control;
	char tab1[24] = { 0 };
	char tab2[1] = { 0 };
	char tab3[7] = { 0 };
	int value;
	
	// Measurement time in ms (Datasheet Appendix B page 26)
	wait_time = 1.25 + (2.3 * OVERSAMPLE_TEMP) + ((2.3 * OVERSAMPLE_PRES) + 0.575) + ((2.3 * OVERSAMPLE_HUM)+0.575);
	
	int fd = openDevice();
	
	// humidity data acquisition (page 26)
	i2c_smbus_write_byte_data(fd, REG_CONTROL_HUM, OVERSAMPLE_HUM);		
	
	// pressure and temperature data acquisition option (page 27)
	control = OVERSAMPLE_TEMP<<5 | OVERSAMPLE_PRES<<2 | MODE;
	i2c_smbus_write_byte_data(fd, REG_CONTROL, control);
	
	// Read Trimming parameter readout
	value =  i2c_smbus_read_i2c_block_data(fd, 0x88, 24, tab1);
	value =  i2c_smbus_read_i2c_block_data(fd, 0xA1, 1, tab2);
	value =  i2c_smbus_read_i2c_block_data(fd, 0xE1, 7, tab3);

	cal.dig_T1 = tab1[1] << 8 | tab1[0] & 0xFF;
	cal.dig_T2 = tab1[3] << 8 | tab1[2] & 0xFF;
	cal.dig_T3 = tab1[5] << 8 | tab1[4] & 0xFF;

	cal.dig_P1 = tab1[7] << 8 | tab1[6] & 0xFF;
	cal.dig_P2 = tab1[9] << 8 | tab1[8] & 0xFF;
	cal.dig_P3 = tab1[11] << 8 | tab1[10] & 0xFF;
	cal.dig_P4 = tab1[13] << 8 | tab1[12] & 0xFF;
	cal.dig_P5 = tab1[15] << 8 | tab1[14] & 0xFF;
	cal.dig_P6 = tab1[17] << 8 | tab1[16] & 0xFF;
	cal.dig_P7 = tab1[19] << 8 | tab1[18] & 0xFF;
	cal.dig_P8 = tab1[21] << 8 | tab1[20] & 0xFF;
	cal.dig_P9 = tab1[23] << 8 | tab1[22] & 0xFF;
	
	cal.dig_H1 = tab2[0];
	cal.dig_H2 = tab3[1] << 8 | tab3[0] & 0xFF;
	cal.dig_H3 = tab3[2];
	cal.dig_H4 = tab3[3] << 4 | (0x0F & tab3[4]);
	cal.dig_H5 = tab3[5] << 4 | ((tab3[6] >> 4 )& 0x0F );
	cal.dig_H6 = tab3[7];

}

static int parse_opts(int argc, char *argv[])
{	
	int long_index = 0;
	int opt;
	
	static struct option option[] = 
	{
		{"temperature", no_argument, 		NULL, 't' },		
		{"humidity", 	no_argument, 		NULL, 'h' },
		{"pressure", 	no_argument, 		NULL, 'p' },
		{"rtfm", 	no_argument, 		NULL, 'r' },
		{0, 			0, 		0, 	0 }
	};
		
 	
	while ((opt = getopt_long(argc, argv, "thpr", option, &long_index)) >= 0) {
		switch(opt) {
			case 'h':
				printf("%.2lf %\n", getHumidity());
				break;
			case 't':
				printf("%.2f °C\n",  getTemperature());
				break;
			case 'p':	
				printf("%.2f hPa\n", getPressure());
				break;
			case 'r':
				display_help(argv[0]);
				exit(EXIT_SUCCESS);	
			default:
				fprintf(stderr,"[ERROR] %s: Bad option. -h for help\n", argv[0]);
				exit(EXIT_FAILURE);
		}				
	}
	
	return 0;
}

int main(int argc, char * argv[])
{
	
	int ret_val = 0;	
  	
  	if (argc < 2){
		display_help(argv[0]);
	}
		
	startCal(); 
		
	ret_val = parse_opts(argc, argv);
	if (ret_val) {
		display_help(argv[0]);
		return ret_val;
	}

	return EXIT_SUCCESS;
}
